@@ -21,6 +21,7 @@
 
 (require 'jabber-chat)
 (require 'jabber-widget)
+(require 'jabber-newdisco)
 
 (require 'cl)
 
@@ -29,6 +30,12 @@
 Keys are strings, the bare JID of the room.
 Values are strings.")
 
+(defvar jabber-pending-groupchats (make-hash-table)
+  "Hash table of groupchats and nicknames.
+Keys are JID symbols; values are strings.
+This table records the last nickname used to join the particular
+chat room.  Items are thus never removed.")
+
 (defvar jabber-muc-participants nil
   "alist of groupchats and participants
 Keys are strings, the bare JID of the room.
@@ -53,6 +60,14 @@ Values are lists of nickname strings.")
   :group 'jabber-chat
   :type '(repeat (string :tag "JID of room")))
 
+(defcustom jabber-muc-disable-disco-check nil
+  "If non-nil, disable checking disco#info of rooms before joining them.
+Disco information can tell whether the room exists and whether it is
+password protected, but some servers do not support it.  If you want
+to join chat rooms on such servers, set this variable to t."
+  :group 'jabber-chat
+  :type 'boolean)
+
 (defcustom jabber-groupchat-buffer-format "*-jabber-groupchat-%n-*"
   "The format specification for the name of groupchat buffers.
 
@@ -116,7 +131,7 @@ The format is that of `mode-line-format' and `header-line-format'."
   :type 'sexp
   :group 'jabber-chat)
 
-(defvar jabber-muc-printers '(jabber-muc-snarf-topic)
+(defvar jabber-muc-printers '()
   "List of functions that may be able to print part of a MUC message.
 This gets prepended to `jabber-chat-printers', which see.")
 
@@ -133,10 +148,11 @@ Either a string or a buffer is returned, so use `get-buffer' or
   "Prepare a buffer for chatroom GROUP.
 This function is idempotent."
   (with-current-buffer (get-buffer-create (jabber-muc-get-buffer group))
-    (if (not (eq major-mode 'jabber-chat-mode)) (jabber-chat-mode))
-    (make-local-variable 'jabber-group)
+    (unless (eq major-mode 'jabber-chat-mode)
+      (jabber-chat-mode #'jabber-chat-pp))
+
+    (set (make-local-variable 'jabber-group) group)
     (make-local-variable 'jabber-muc-topic)
-    (setq jabber-group group)
     (setq jabber-send-function 'jabber-muc-send)
     (setq header-line-format jabber-muc-header-line-format)
     (current-buffer)))
@@ -154,9 +170,10 @@ Either a string or a buffer is returned, so use `get-buffer' or
   "Prepare a buffer for chatting with NICKNAME in GROUP.
 This function is idempotent."
   (with-current-buffer (get-buffer-create (jabber-muc-private-get-buffer group nickname))
-    (if (not (eq major-mode 'jabber-chat-mode)) (jabber-chat-mode))
-    (make-local-variable 'jabber-chatting-with)
-    (setq jabber-chatting-with (concat group "/" nickname))
+    (unless (eq major-mode 'jabber-chat-mode)
+      (jabber-chat-mode #'jabber-chat-pp))
+
+    (set (make-local-variable 'jabber-chatting-with) (concat group "/" nickname))
     (setq jabber-send-function 'jabber-chat-send)
     (setq header-line-format jabber-muc-private-header-line-format)
 
@@ -169,7 +186,7 @@ This function is idempotent."
   (jabber-send-sexp `(message
 		      ((to . ,jabber-group)
 		       (type . "groupchat"))
-		      (body () ,(jabber-escape-xml body)))))
+		      (body () ,body))))
 
 (defun jabber-muc-add-groupchat (group nickname)
   "Remember participating in GROUP under NICKNAME."
@@ -203,12 +220,82 @@ Return nil if nothing known about that combination."
 	  ;; and maybe this participant is already in the list
 	  (if participant
 	      ;; if so, just update role, affiliation, etc.
-	      ;; XXX: calculate delta and report to user? e.g. "X was given voice"
 	      (setf (cdr participant) new-plist)
 	    (push (cons nickname new-plist) (cdr participants))))
       ;; or we don't
       (push (cons group (list (cons nickname new-plist))) jabber-muc-participants))))
 
+(defun jabber-muc-report-delta (nickname old-plist new-plist reason actor)
+  "Compare OLD-PLIST and NEW-PLIST, and return a string explaining the change.
+Return nil if nothing noteworthy has happened.
+NICKNAME is the user experiencing the change.  REASON and ACTOR, if non-nil,
+are the corresponding presence fields.
+
+This function is only concerned with presence stanzas resulting
+in the user entering/staying in the room."
+  ;; The keys in the plist are affiliation, role and jid.
+  (cond
+   ((null old-plist)
+    ;; User enters the room
+    (concat nickname " enters the room ("
+	    (plist-get new-plist 'role)
+	    (unless (string= (plist-get new-plist 'affiliation) "none")
+	      (concat ", " (plist-get new-plist 'affiliation)))
+	    ")"))
+
+   ;; If affiliation changes, the role change is usually the logical
+   ;; one, so don't report it separately.
+   ((not (string= (plist-get old-plist 'affiliation)
+		  (plist-get new-plist 'affiliation)))
+    (let ((actor-reason (concat (when actor
+				  (concat " by " actor))
+				(when reason
+				  (concat ": " reason))))
+	  (from (plist-get old-plist 'affiliation))
+	  (to (plist-get new-plist 'affiliation)))
+      ;; There are many ways to express these transitions in English.
+      ;; This one favors eloquence over regularity and consistency.
+      (cond
+       ;; Higher affiliation
+       ((or (and (member from '("outcast" "none" "member"))
+		 (member to '("admin" "owner")))
+	    (and (string= from "admin") (string= to "owner")))
+	(concat nickname " has been promoted to " to actor-reason))
+       ;; Lower affiliation
+       ((or (and (member from '("owner" "admin"))
+		 (string= to "member"))
+	    (and (string= from "owner") (string= to "admin")))
+	(concat nickname " has been demoted to " to actor-reason))
+       ;; Become member
+       ((string= to "member")
+	(concat nickname " has been granted membership" actor-reason))
+       ;; Lose membership
+       ((string= to "none")
+	(concat nickname " has been deprived of membership" actor-reason)))))
+
+   ;; Role changes
+   ((not (string= (plist-get old-plist 'role)
+		  (plist-get new-plist 'role)))
+    (let ((actor-reason (concat (when actor
+				  (concat " by " actor))
+				(when reason
+				  (concat ": " reason))))
+	  (from (plist-get old-plist 'role))
+	  (to (plist-get new-plist 'role)))
+      ;; Possible roles are "none" (not in room, hence not of interest
+      ;; in this function), "visitor" (no voice), "participant" (has
+      ;; voice), and "moderator".
+      (cond
+       ((string= to "moderator")
+	(concat nickname " has been granted moderator privileges" actor-reason))
+       ((and (string= from "moderator")
+	     (string= to "participant"))
+	(concat nickname " had moderator privileges revoked" actor-reason))
+       ((string= to "participant")
+	(concat nickname " has been granted voice" actor-reason))
+       ((string= to "visitor")
+	(concat nickname " has been denied voice" actor-reason)))))))
+
 (defun jabber-muc-remove-participant (group nickname)
   "Forget everything about NICKNAME in GROUP."
   (let ((participants (assoc group jabber-muc-participants)))
@@ -216,15 +303,17 @@ Return nil if nothing known about that combination."
       (let ((participant (assoc nickname (cdr participants))))
 	(setf (cdr participants) (delq participant (cdr participants)))))))
 
-(defun jabber-muc-read-completing (prompt)
-  "Read the name of a joined chatroom, or use chatroom of current buffer, if any."
+(defun jabber-muc-read-completing (prompt &optional allow-not-joined)
+  "Read the name of a joined chatroom, or use chatroom of current buffer, if any.
+If ALLOW-NOT-JOINED is provided and true, permit choosing any
+JID; only provide completion as a guide."
   (or jabber-group
       (jabber-read-jid-completing prompt
 				  (if (null *jabber-active-groupchats*)
 				      (error "You haven't joined any group")
 				    (mapcar (lambda (x) (jabber-jid-symbol (car x)))
 					    *jabber-active-groupchats*))
-				  t
+				  (not allow-not-joined)
 				  jabber-group)))
 
 (defun jabber-muc-read-nickname (group prompt)
@@ -238,7 +327,7 @@ Return nil if nothing known about that combination."
    (cons "Configure groupchat" 'jabber-groupchat-get-config))
 (defun jabber-groupchat-get-config (group)
   "Ask for MUC configuration form"
-  (interactive (list (jabber-muc-read-completing "Configure group: ")))
+  (interactive (list (jabber-muc-read-completing "Configure group: " t)))
   (jabber-send-iq group
 		  "get"
 		  '(query ((xmlns . "http://jabber.org/protocol/muc#owner")))
@@ -290,24 +379,80 @@ Return nil if nothing known about that combination."
 (add-to-list 'jabber-jid-muc-menu
 	     (cons "Join groupchat" 'jabber-groupchat-join))
 
-(defun jabber-groupchat-join (group nickname)
-  "join a groupchat, or change nick"
+(defun jabber-groupchat-join (group nickname &optional popup)
+  "join a groupchat, or change nick.
+In interactive calls, or if POPUP is true, switch to the
+groupchat buffer."
   (interactive 
    (let ((group (jabber-read-jid-completing "group: ")))
-     (list group (jabber-muc-read-my-nickname group))))
+     (list group (jabber-muc-read-my-nickname group) t)))
+
+  ;; If the user is already in the room, we don't need as many checks.
+  (if (or (assoc group *jabber-active-groupchats*)
+	  ;; Or if the users asked us not to check disco info.
+	  jabber-muc-disable-disco-check)
+      (jabber-groupchat-join-3 group nickname nil popup)
+    ;; Else, send a disco request to find out what we are connecting
+    ;; to.
+    (jabber-disco-get-info group nil #'jabber-groupchat-join-2
+			   (list group nickname popup))))
+
+(defun jabber-groupchat-join-2 (closure result)
+  (destructuring-bind (group nickname popup) closure
+    (let ( ;; Either success...
+	  (identities (car result))
+	  (features (cadr result))
+	  ;; ...or error
+	  (condition (when (eq (car result) 'error) (jabber-error-condition result))))
+      (cond
+       ;; Maybe the room doesn't exist yet.
+       ((eq condition 'item-not-found)
+	(unless (y-or-n-p (format "%s doesn't exist.  Create it? " (jabber-jid-displayname group)))
+	  (error "Non-existent groupchat")))
+
+       ;; Maybe the room doesn't support disco.
+       ((eq condition 'feature-not-implemented)
+	t				;whatever...
+	)
+       ;; Maybe another error occurred.
+       (condition
+	(error "Couldn't query groupchat: %s" (jabber-parse-error result)))
+
+       ;; Maybe it isn't a chat room.
+       ((not (find "conference" identities 
+		   :key (lambda (i) (aref i 1))
+		   :test #'string=))
+	(error "%s is not a groupchat" (jabber-jid-displayname group))))
+
+      (let ((password
+	     ;; Is the room password-protected?
+	     (when (member "muc_passwordprotected" features)
+	       (read-passwd (format "Password for %s: " (jabber-jid-displayname group))))))
+
+	(jabber-groupchat-join-3 group nickname password popup)))))
+
+(defun jabber-groupchat-join-3 (group nickname password popup)
+
   ;; Remember that this is a groupchat _before_ sending the stanza.
   ;; The response might come quicker than you think.
-  (let ((whichgroup (assoc group *jabber-active-groupchats*)))
-    (if whichgroup
-	(setcdr whichgroup nickname)
-      (add-to-list '*jabber-active-groupchats* (cons group nickname))))
+
+  (puthash (jabber-jid-symbol group) nickname jabber-pending-groupchats)
   
   (jabber-send-sexp `(presence ((to . ,(format "%s/%s" group nickname)))
-			       (x ((xmlns . "http://jabber.org/protocol/muc")))))
+			       (x ((xmlns . "http://jabber.org/protocol/muc"))
+				  ,@(when password
+				      `((password () ,password))))
+			       ,@(jabber-presence-children)))
 
-  (let ((buffer (jabber-muc-create-buffer group)))
-    ;; We don't want to switch to autojoined groupchats
-    (when (interactive-p)
+  ;; There, stanza sent.  Now we just wait for the MUC service to
+  ;; mirror the stanza.  This is handled in
+  ;; `jabber-muc-process-presence', where a buffer will be created for
+  ;; the room.
+
+  ;; But if the user interactively asked to join, he/she probably
+  ;; wants the buffer to pop up right now.
+  (when popup
+    (let ((buffer (jabber-muc-create-buffer group)))
       (switch-to-buffer buffer))))
 
 (defun jabber-muc-read-my-nickname (group)
@@ -342,14 +487,19 @@ Return nil if nothing known about that combination."
   "Print names, affiliations, and roles of participants in GROUP."
   (interactive (list (jabber-muc-read-completing "Group: ")))
   (with-current-buffer (jabber-muc-create-buffer group)
-    (let ((jabber-chat-fill-long-lines nil))
-      (jabber-chat-buffer-display 'jabber-muc-system-prompt nil
-				  '(jabber-muc-print-names)
-				  (cdr (assoc group jabber-muc-participants))))))
+    (ewoc-enter-last jabber-chat-ewoc (list :notice
+					    (jabber-muc-print-names
+					     (cdr (assoc group jabber-muc-participants)))
+					    :time (current-time)))
+    ;; (let ((jabber-chat-fill-long-lines nil))
+;;       (jabber-chat-buffer-display 'jabber-muc-system-prompt nil
+;; 				  '(jabber-muc-print-names)
+;; 				  (cdr (assoc group jabber-muc-participants))))
+    ))
 
 (defun jabber-muc-print-names (participants)
-  "Format and insert data in PARTICIPANTS."
-  (apply 'insert "Participants:\n"
+  "Format and return data in PARTICIPANTS."
+  (apply 'concat "Participants:\n"
 	 (format "%-15s %-15s %-11s %s\n" "Nickname" "Role" "Affiliation" "JID")
 	 (mapcar (lambda (x)
 		   (let ((plist (cdr x)))
@@ -455,63 +605,64 @@ group, else it is a JID."
 
 (add-to-list 'jabber-body-printers 'jabber-muc-print-invite)
 
-(defun jabber-muc-print-invite (xml-data)
+(defun jabber-muc-print-invite (xml-data who mode)
   "Print MUC invitation"
   (dolist (x (jabber-xml-get-children xml-data 'x))
     (when (string= (jabber-xml-get-attribute x 'xmlns) "http://jabber.org/protocol/muc#user")
       (let ((invitation (car (jabber-xml-get-children x 'invite))))
 	(when invitation
-	  (let ((group (jabber-xml-get-attribute xml-data 'from))
-		(inviter (jabber-xml-get-attribute invitation 'from))
-		(reason (car (jabber-xml-node-children (car (jabber-xml-get-children invitation 'reason))))))
-	    ;; XXX: password
-	    (insert "You have been invited to MUC room " (jabber-jid-displayname group))
-	    (when inviter
-	      (insert " by " (jabber-jid-displayname inviter)))
-	    (insert ".")
-	    (when reason
-	      (insert "  Reason: " reason))
-	    (insert "\n\n")
+	  (when (eql mode :insert)
+	    (let ((group (jabber-xml-get-attribute xml-data 'from))
+		  (inviter (jabber-xml-get-attribute invitation 'from))
+		  (reason (car (jabber-xml-node-children (car (jabber-xml-get-children invitation 'reason))))))
+	      ;; XXX: password
+	      (insert "You have been invited to MUC room " (jabber-jid-displayname group))
+	      (when inviter
+		(insert " by " (jabber-jid-displayname inviter)))
+	      (insert ".")
+	      (when reason
+		(insert "  Reason: " reason))
+	      (insert "\n\n")
 
-	    (let ((action
-		   `(lambda (&rest ignore) (interactive)
-		      (jabber-groupchat-join ,group
-					     (jabber-muc-read-my-nickname ,group)))))
-	      (if (fboundp 'insert-button)
-		  (insert-button "Accept"
-				 'action action)
-	      ;; Simple button replacement
-	      (let ((keymap (make-keymap)))
-		(define-key keymap "\r" action)
-		(insert (jabber-propertize "Accept"
-					   'keymap keymap
-					   'face 'highlight))))
+	      (let ((action
+		     `(lambda (&rest ignore) (interactive)
+			(jabber-groupchat-join ,group
+					       (jabber-muc-read-my-nickname ,group)))))
+		(if (fboundp 'insert-button)
+		    (insert-button "Accept"
+				   'action action)
+		  ;; Simple button replacement
+		  (let ((keymap (make-keymap)))
+		    (define-key keymap "\r" action)
+		    (insert (jabber-propertize "Accept"
+					       'keymap keymap
+					       'face 'highlight))))
 
-	    (insert "\t")
+		(insert "\t")
 
-	    (let ((action
-		   `(lambda (&rest ignore) (interactive)
-		      (let ((reason
-			     (jabber-read-with-input-method
-			      "Reason: ")))
-			(jabber-send-sexp
-			 (list 'message
-			       (list (cons 'to ,group))
-			       (list 'x
-				     (list (cons 'xmlns "http://jabber.org/protocol/muc#user"))
-				     (list 'decline
-					   (list (cons 'to ,inviter))
-					   (unless (zerop (length reason))
-					     (list 'reason nil reason))))))))))
-	      (if (fboundp 'insert-button)
-		  (insert-button "Decline"
-				 'action action)
-		;; Simple button replacement
-		(let ((keymap (make-keymap)))
-		  (define-key keymap "\r" action)
-		  (insert (jabber-propertize "Decline"
-					     'keymap keymap
-					     'face 'highlight)))))))
+		(let ((action
+		       `(lambda (&rest ignore) (interactive)
+			  (let ((reason
+				 (jabber-read-with-input-method
+				  "Reason: ")))
+			    (jabber-send-sexp
+			     (list 'message
+				   (list (cons 'to ,group))
+				   (list 'x
+					 (list (cons 'xmlns "http://jabber.org/protocol/muc#user"))
+					 (list 'decline
+					       (list (cons 'to ,inviter))
+					       (unless (zerop (length reason))
+						 (list 'reason nil reason))))))))))
+		  (if (fboundp 'insert-button)
+		      (insert-button "Decline"
+				     'action action)
+		    ;; Simple button replacement
+		    (let ((keymap (make-keymap)))
+		      (define-key keymap "\r" action)
+		      (insert (jabber-propertize "Decline"
+						 'keymap keymap
+						 'face 'highlight))))))))
 	  (return t))))))
 
 (defun jabber-muc-autojoin ()
@@ -524,7 +675,8 @@ group, else it is a JID."
 
 (defun jabber-muc-message-p (message)
   "Return non-nil if MESSAGE is a groupchat message.
-That does not include private messages in a groupchat."
+That does not include private messages in a groupchat, but does
+include groupchat invites."
   ;; Public groupchat messages have type "groupchat" and are from
   ;; room@server/nick.  Public groupchat errors have type "error" and
   ;; are from room@server.
@@ -533,7 +685,8 @@ That does not include private messages in a groupchat."
     (or 
      (string= type "groupchat")
      (and (string= type "error")
-	  (assoc from *jabber-active-groupchats*)))))
+	  (gethash (jabber-jid-symbol from) jabber-pending-groupchats))
+     (jabber-xml-path message '(("http://jabber.org/protocol/muc#user" . "x") invite)))))
 
 (defun jabber-muc-sender-p (jid)
   "Return non-nil if JID is a full JID of an MUC participant."
@@ -561,8 +714,17 @@ That does not include private messages in a groupchat."
 
 (defun jabber-muc-presence-p (presence)
   "Return non-nil if PRESENCE is presence from groupchat."
-  (let ((from (jabber-xml-get-attribute presence 'from)))
-    (assoc (jabber-jid-user from) *jabber-active-groupchats*)))
+  (let ((from (jabber-xml-get-attribute presence 'from))
+	(type (jabber-xml-get-attribute presence 'type))
+	(muc-marker (find-if 
+		     (lambda (x) (equal (jabber-xml-get-attribute x 'xmlns)
+				   "http://jabber.org/protocol/muc#user"))
+		     (jabber-xml-get-children presence 'x))))
+    ;; This is MUC presence if it has an MUC-namespaced tag...
+    (or muc-marker
+	;; ...or if it is error presence from a room we tried to join.
+	(and (string= type "error")
+	     (gethash (jabber-jid-symbol from) jabber-pending-groupchats)))))
 
 (defun jabber-muc-parse-affiliation (x-muc)
   "Parse X-MUC in the muc#user namespace and return a plist.
@@ -572,11 +734,10 @@ Return nil if X-MUC is nil."
 			(jabber-xml-node-attributes
 			 (car (jabber-xml-get-children x-muc 'item))))))
 
-(defun jabber-muc-print-prompt (xml-data)
+(defun jabber-muc-print-prompt (xml-data &optional local)
   "Print MUC prompt for message in XML-DATA."
   (let ((nick (jabber-jid-resource (jabber-xml-get-attribute xml-data 'from)))
 	(timestamp (car (delq nil (mapcar 'jabber-x-delay (jabber-xml-get-children xml-data 'x))))))
-    (jabber-maybe-print-rare-time timestamp)
     (if (stringp nick)
 	(insert (jabber-propertize
 		 (format-spec jabber-groupchat-prompt-format
@@ -590,7 +751,8 @@ Return nil if X-MUC is nil."
 			       (cons ?u nick)
 			       (cons ?r nick)
 			       (cons ?j (concat jabber-group "/" nick))))
-		 'face 'jabber-chat-prompt-foreign
+		 'face (if local 'jabber-chat-prompt-local
+                         'jabber-chat-prompt-foreign)
 		 'help-echo (concat (format-time-string "On %Y-%m-%d %H:%M:%S" timestamp) " from " nick " in " jabber-group)))
       (jabber-muc-system-prompt))))
 
@@ -599,7 +761,6 @@ Return nil if X-MUC is nil."
   (let ((nick (jabber-jid-resource (jabber-xml-get-attribute xml-data 'from)))
 	(group (jabber-jid-user (jabber-xml-get-attribute xml-data 'from)))
 	(timestamp (car (delq nil (mapcar 'jabber-x-delay (jabber-xml-get-children xml-data 'x))))))
-    (jabber-maybe-print-rare-time timestamp)
     (insert (jabber-propertize
 	     (format-spec jabber-muc-private-foreign-prompt-format
 			  (list
@@ -616,7 +777,6 @@ Return nil if X-MUC is nil."
 
 (defun jabber-muc-system-prompt (&rest ignore)
   "Print system prompt for MUC."
-  (jabber-maybe-print-rare-time nil)
   (insert (jabber-propertize
 	   (format-spec jabber-groupchat-prompt-format
 			(list
@@ -637,19 +797,24 @@ Return nil if X-MUC is nil."
 	   (group (jabber-jid-user from))
 	   (nick (jabber-jid-resource from))
 	   (error-p (jabber-xml-get-children xml-data 'error))
+	   (type (cond 
+		  (error-p :muc-error)
+		  ((string= nick (cdr (assoc group *jabber-active-groupchats*)))
+		   :muc-local)
+		  (t :muc-foreign)))
 	   (body-text (car (jabber-xml-node-children
 			   (car (jabber-xml-get-children
-				 xml-data 'body))))))
+				 xml-data 'body)))))
+
+	   (printers (append jabber-muc-printers jabber-chat-printers)))
+
       (with-current-buffer (jabber-muc-create-buffer group)
+	(jabber-muc-snarf-topic xml-data)
 	;; Call alert hooks only when something is output
-	(when
-	    (jabber-chat-buffer-display 'jabber-muc-print-prompt
-					xml-data
-					(if error-p
-					    '(jabber-chat-print-error)
-					  (append jabber-muc-printers
-						  jabber-chat-printers))
-					xml-data)
+	(when (or error-p
+		  (run-hook-with-args-until-success 'printers xml-data type :printp))
+	  (jabber-maybe-print-rare-time
+	   (ewoc-enter-last jabber-chat-ewoc (list type xml-data :time (current-time))))
 	  
 	  (dolist (hook '(jabber-muc-hooks jabber-alert-muc-hooks))
 	    (run-hook-with-args hook
@@ -659,86 +824,103 @@ Return nil if X-MUC is nil."
 
 (defun jabber-muc-process-presence (presence)
   (let* ((from (jabber-xml-get-attribute presence 'from))
-	(type (jabber-xml-get-attribute presence 'type))
-	(x-muc (find-if 
-		(lambda (x) (equal (jabber-xml-get-attribute x 'xmlns)
-				   "http://jabber.org/protocol/muc#user"))
-		(jabber-xml-get-children presence 'x)))
-	(group (jabber-jid-user from))
-	(nickname (jabber-jid-resource from))
-	(symbol (jabber-jid-symbol from))
-	(item (car (jabber-xml-get-children x-muc 'item)))
-	(actor (jabber-xml-get-attribute (car (jabber-xml-get-children item 'actor)) 'jid))
-	(reason (car (jabber-xml-node-children (car (jabber-xml-get-children item 'reason)))))
-	(status-code (jabber-xml-get-attribute
-		      (car (jabber-xml-get-children x-muc 'status))
-		      'code))
-	(error-node (car (jabber-xml-get-children presence 'error))))
+	 (type (jabber-xml-get-attribute presence 'type))
+	 (x-muc (find-if 
+		 (lambda (x) (equal (jabber-xml-get-attribute x 'xmlns)
+			       "http://jabber.org/protocol/muc#user"))
+		 (jabber-xml-get-children presence 'x)))
+	 (group (jabber-jid-user from))
+	 (nickname (jabber-jid-resource from))
+	 (symbol (jabber-jid-symbol from))
+	 (item (car (jabber-xml-get-children x-muc 'item)))
+	 (actor (jabber-xml-get-attribute (car (jabber-xml-get-children item 'actor)) 'jid))
+	 (reason (car (jabber-xml-node-children (car (jabber-xml-get-children item 'reason)))))
+	 (status-code (jabber-xml-get-attribute
+		       (car (jabber-xml-get-children x-muc 'status))
+		       'code))
+	 (error-node (car (jabber-xml-get-children presence 'error))))
     ;; handle leaving a room
     (cond 
      ((or (string= type "unavailable") (string= type "error"))
       ;; are we leaving?
-      (if (string= nickname (cdr (assoc group *jabber-active-groupchats*)))
-	  (progn
+      (if (string= nickname (gethash (jabber-jid-symbol group) jabber-pending-groupchats))
+	  (let ((message (cond
+			  ((string= type "error")
+			   (concat "Error entering room"
+				   (when error-node
+				     (concat ": " (jabber-parse-error error-node)))))
+			  ((equal status-code "301")
+			   (concat "You have been banned"
+				   (when actor (concat " by " actor))
+				   (when reason (concat " - '" reason "'"))))
+			  ((equal status-code "307")
+			   (concat "You have been kicked"
+				   (when actor (concat " by " actor))
+				   (when reason (concat " - '" reason "'"))))
+			  (t
+			   "You have left the chatroom"))))
 	    (jabber-muc-remove-groupchat group)
 	    ;; If there is no buffer for this groupchat, don't bother
 	    ;; creating one just to tell that user left the room.
 	    (let ((buffer (get-buffer (jabber-muc-get-buffer group))))
-	      (when buffer
-		(with-current-buffer buffer
-		  (jabber-chat-buffer-display 
-		   'jabber-muc-system-prompt
-		   nil
-		   '(insert)
-		   (cond
-		    ((string= type "error")
-		     (jabber-propertize
-		      (concat "Error entering room"
-			      (when error-node
-				(concat ": " (jabber-parse-error error-node))))
-		      'face 'jabber-chat-error))
-		    ((equal status-code "301")
-		     (concat "You have been banned"
-			     (when actor (concat " by " actor))
-			     (when reason (concat " - '" reason "'"))))
-		    ((equal status-code "307")
-		     (concat "You have been kicked"
-			     (when actor (concat " by " actor))
-			     (when reason (concat " - '" reason "'"))))
-		    (t
-		     "You have left the chatroom")))))))
+	      (if buffer
+		  (with-current-buffer buffer
+		    (jabber-maybe-print-rare-time
+		     (ewoc-enter-last jabber-chat-ewoc
+				      (list (if (string= type "error")
+						:muc-error
+					      :muc-notice)
+					    message
+					    :time (current-time)))))
+		(message "%s: %s" (jabber-jid-displayname group) message))))
 	;; or someone else?
 	(jabber-muc-remove-participant group nickname)
 	(with-current-buffer (jabber-muc-create-buffer group)
-	  (jabber-chat-buffer-display 
-	   'jabber-muc-system-prompt
-	   nil
-	   '(insert)
-	   (cond
-	    ((equal status-code "301")
-	     (concat nickname " has been banned"
-		     (when actor (concat " by " actor))
-		     (when reason (concat " - '" reason "'"))))
-	    ((equal status-code "307")
-	     (concat nickname " has been kicked"
-		     (when actor (concat " by " actor))
-		     (when reason (concat " - '" reason "'"))))
-	    ((equal status-code "303")
-	     (concat nickname " changes nickname to "
-		     (jabber-xml-get-attribute item 'nick)))
-	    (t
-	     (concat nickname " has left the chatroom")))))))
+	  (jabber-maybe-print-rare-time
+	   (ewoc-enter-last
+	    jabber-chat-ewoc
+	    (list :muc-notice
+		  (cond
+		   ((equal status-code "301")
+		    (concat nickname " has been banned"
+			    (when actor (concat " by " actor))
+			    (when reason (concat " - '" reason "'"))))
+		   ((equal status-code "307")
+		    (concat nickname " has been kicked"
+			    (when actor (concat " by " actor))
+			    (when reason (concat " - '" reason "'"))))
+		   ((equal status-code "303")
+		    (concat nickname " changes nickname to "
+			    (jabber-xml-get-attribute item 'nick)))
+		   (t
+		    (concat nickname " has left the chatroom")))
+		  :time (current-time)))))))
      (t 
       ;; someone is entering
-      (let ((new-participant (not (jabber-muc-participant-plist group nickname)))
+
+      (when (string= nickname (gethash (jabber-jid-symbol group) jabber-pending-groupchats))
+	;; Our own nick?  We just succeeded in entering the room.
+	(let ((whichgroup (assoc group *jabber-active-groupchats*)))
+	  (if whichgroup
+	      (setcdr whichgroup nickname)
+	    (add-to-list '*jabber-active-groupchats* (cons group nickname)))))	
+
+      ;; Whoever enters, we create a buffer (if it didn't already
+      ;; exist), and print a notice.  This is where autojoined MUC
+      ;; rooms have buffers created for them.  We also remember some
+      ;; metadata.
+      (let ((old-plist (jabber-muc-participant-plist group nickname))
 	    (new-plist (jabber-muc-parse-affiliation x-muc)))
 	(jabber-muc-modify-participant group nickname new-plist)
-	(when new-participant
-	  (with-current-buffer (jabber-muc-create-buffer group)
-	    (jabber-chat-buffer-display 'jabber-muc-system-prompt
-					nil
-					'(insert)
-					(format "%s enters the chatroom" nickname)))))))))
+	(let ((report (jabber-muc-report-delta nickname old-plist new-plist
+					       reason actor)))
+	  (when report
+	    (with-current-buffer (jabber-muc-create-buffer group)
+	      (jabber-maybe-print-rare-time
+	       (ewoc-enter-last
+		jabber-chat-ewoc
+		(list :muc-notice report
+		      :time (current-time))))))))))))
 	      
 (provide 'jabber-muc)
 
