@@ -28,11 +28,12 @@
 (defgroup jabber-roster nil "roster display options"
   :group 'jabber)
 
-(defcustom jabber-roster-line-format " %c %-25n %u %-8s  %S\n"
+(defcustom jabber-roster-line-format " %a %c %-25n %u %-8s  %S\n"
   "The format specification of the lines in the roster display.
 
 These fields are available:
 
+%a   Avatar, if any
 %c   \"*\" if the contact is connected, or \" \" if not
 %u   sUbscription state - see below
 %n   Nickname of contact, or JID if no nickname
@@ -173,6 +174,9 @@ Trailing newlines are always removed, regardless of this variable."
     (set-keymap-parent map jabber-common-keymap)
     (define-key map [mouse-2] 'jabber-popup-combined-menu)
     (define-key map (kbd "TAB") 'jabber-go-to-next-jid)
+    (define-key map (kbd "S-TAB") 'jabber-go-to-previous-jid)
+    (define-key map (kbd "M-TAB") 'jabber-go-to-previous-jid)
+    (define-key map (kbd "<backtab>") 'jabber-go-to-previous-jid)
     (define-key map (kbd "RET") 'jabber-chat-with-jid-at-point)
     (define-key map (kbd "C-k") 'jabber-roster-delete-jid-at-point)
 
@@ -186,6 +190,7 @@ Trailing newlines are always removed, regardless of this variable."
     (define-key map "v" 'jabber-get-version)
     (define-key map "a" 'jabber-send-presence)
     (define-key map "g" 'jabber-display-roster)
+    (define-key map "S" 'jabber-ft-send)
     ;;(define-key map "D" 'jabber-disconnect)
     map))
 
@@ -243,19 +248,22 @@ See `jabber-sort-order' for order used."
 
 (defun jabber-roster-sort-by-displayname (a b)
   "Sort roster items by displayed name."
-  (cond
-   ((string-lessp a b) -1)
-   ((string= a b) 0)
-   (t 1)))
+  (let ((a-name (jabber-jid-displayname a))
+	(b-name (jabber-jid-displayname b)))
+    (cond
+     ((string-lessp a-name b-name) -1)
+     ((string= a-name b-name) 0)
+     (t 1))))
 
 (defun jabber-fix-status (status)
   "Make status strings more readable"
-  (when (string-match "\n+$" status)
-    (setq status (replace-match "" t t status)))
-  (when jabber-remove-newlines
-    (while (string-match "\n" status)
-      (setq status (replace-match " " t t status))))
-  status)
+  (when status
+    (when (string-match "\n+$" status)
+      (setq status (replace-match "" t t status)))
+    (when jabber-remove-newlines
+      (while (string-match "\n" status)
+	(setq status (replace-match " " t t status))))
+    status))
 
 (defvar jabber-roster-positions nil
   "Alist tracking positions of items in the roster.
@@ -330,9 +338,12 @@ C-c C-s  Service menu
 	(move-to-column current-column)))))
 
 (defun jabber-display-roster-entry (buddy)
-  "Format and insert a roster entry for BUDDY at point."
+  "Format and insert a roster entry for BUDDY at point.
+BUDDY is a JID symbol."
   (let ((buddy-str (format-spec jabber-roster-line-format
 				(list 
+				 (cons ?a (jabber-propertize " "
+							     'display (get buddy 'avatar)))
 				 (cons ?c (if (get buddy 'connected) "*" " "))
 				 (cons ?u (cdr (assoc (or (get buddy 'subscription) "none")
 						      jabber-roster-subscription-display)))
@@ -405,7 +416,8 @@ C-c C-s  Service menu
 	    (insert resource-str)))))))
 
 (defun jabber-presence-update-roster (who &rest ignore)
-  "Update roster without redrawing all of it, if possible."
+  "Update roster without redrawing all of it, if possible.
+WHO is a JID symbol."
   
   (let* ((bare-jid (jabber-jid-symbol 
 		    (jabber-jid-user
@@ -451,6 +463,18 @@ C-c C-s  Service menu
     (if next (goto-char (1+ next))
       (goto-char (point-min)))))
 
+(defun jabber-go-to-previous-jid ()
+  "Move the cursor to the previous jid in the buffer"
+  (interactive)
+  (let ((previous (previous-single-property-change (point) 'jabber-jid)))
+    (when (and previous
+               (not (get-text-property previous 'jabber-jid)))
+      (setq previous (previous-single-property-change previous 'jabber-jid)))
+    (unless previous
+      (setq previous (previous-single-property-change (point-max) 'jabber-jid)))
+    (if previous (goto-char previous)
+      (goto-char (point-max)))))
+
 (provide 'jabber-roster)
 
 ;;; arch-tag: 096af063-0526-4dd2-90fd-bc6b5ba07d32
