@@ -1,4 +1,4 @@
-;; jabber-util.el - various utility functions
+;; jabber-util.el - various utility functions    -*- coding: utf-8; -*-
 
 ;; Copyright (C) 2002, 2003, 2004 - tom berger - object@intelectronica.net
 ;; Copyright (C) 2003, 2004 - Magnus Henoch - mange@freemail.hu
@@ -19,12 +19,11 @@
 ;; along with this program; if not, write to the Free Software
 ;; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
+(eval-when-compile (require 'cl))
+
 (defvar jabber-jid-history nil
   "History of entered JIDs")
 
-(defvar *jabber-sound-playing* nil
-  "is a sound playing right now?")
-
 (cond
  ((fboundp 'replace-in-string)
   (defsubst jabber-replace-in-string (str regexp newtext)
@@ -140,13 +139,18 @@ properties to add to the result."
   (equal (jabber-jid-user jid)
 	 (concat jabber-username "@" jabber-server)))
 
-(defun jabber-read-jid-completing (prompt &optional subset require-match default)
+(defun jabber-read-jid-completing (prompt &optional subset require-match default resource)
   "read a jid out of the current roster from the minibuffer.
 If SUBSET is non-nil, it should be a list of symbols from which
 the JID is to be selected, instead of using the entire roster.
 If REQUIRE-MATCH is non-nil, the JID must be in the list used.
 If DEFAULT is non-nil, it's used as the default value, otherwise
-the default is inferred from context."
+the default is inferred from context.
+RESOURCE is one of the following:
+
+nil         Accept full or bare JID, as entered
+full        Turn bare JIDs to full ones with highest-priority resource
+bare-or-muc Turn full JIDs to bare ones, except for in MUC"
   (let ((jid-at-point (or 
 		       (and default
 			    ;; default can be either a symbol or a string
@@ -159,7 +163,8 @@ the default is inferred from context."
 	(completion-ignore-case t)
 	(jid-completion-table (mapcar #'(lambda (item)
 					  (cons (symbol-name item) item))
-				      (or subset *jabber-roster*))))
+				      (or subset *jabber-roster*)))
+	chosen)
     (dolist (item (or subset *jabber-roster*))
       (if (get item 'name)
 	  (push (cons (get item 'name) item) jid-completion-table)))
@@ -172,10 +177,32 @@ the default is inferred from context."
 					(format "(default %s) " jid-at-point)))
 			    jid-completion-table
 			    nil require-match nil 'jabber-jid-history jid-at-point)))
-      (if (and input (assoc-ignore-case input jid-completion-table))
-	  (symbol-name (cdr (assoc-ignore-case input jid-completion-table)))
-	(and (not (zerop (length input)))
-	     input)))))
+      (setq chosen
+	    (if (and input (assoc-ignore-case input jid-completion-table))
+		(symbol-name (cdr (assoc-ignore-case input jid-completion-table)))
+	      (and (not (zerop (length input)))
+		   input))))
+
+    (when chosen
+      (case resource
+	(full
+	 ;; If JID is bare, add the highest-priority resource.
+	 (if (jabber-jid-resource chosen)
+	     chosen
+	   (let ((highest-resource (get (jabber-jid-symbol chosen) 'resource)))
+	     (if highest-resource
+		 (concat chosen "/" highest-resource)
+	       chosen))))
+	(bare-or-muc
+	 ;; If JID is full and non-MUC, remove resource.
+	 (if (null (jabber-jid-resource chosen))
+	     chosen
+	   (let ((bare (jabber-jid-user chosen)))
+	     (if (assoc bare *jabber-active-groupchats*)
+		 chosen
+	       bare))))
+	(t
+	 chosen)))))
 
 (defun jabber-read-node (prompt)
   "Read node name, taking default from disco item at point."
@@ -188,7 +215,12 @@ the default is inferred from context."
 (defun jabber-read-passwd (&optional prompt)
   "Read Jabber password, either from customized variable or from minibuffer.
 See `jabber-password'."
-  (or jabber-password (read-passwd (or prompt "Jabber password: "))))
+  (if jabber-password
+      ;; Need to copy the password, as sasl.el wants to erase it.  The
+      ;; variable jabber-password is a high-convenience low-security
+      ;; alternative anyway.
+      (copy-sequence jabber-password)
+    (read-passwd (or prompt "Jabber password: "))))
 
 (defun jabber-iq-query (xml-data)
   "Return the query part of an IQ stanza.
@@ -358,6 +390,15 @@ See secton 9.3, Stanza Errors, of XMPP Core, and JEP-0086, Legacy Errors."
     (concat condition
 	    (if text (format ": %s" text)))))
 
+(defun jabber-error-condition (error-xml)
+  "Parse the given <error/> tag and return the condition symbol."
+  (catch 'condition
+    (dolist (child (jabber-xml-node-children error-xml))
+      (when (string=
+		 (jabber-xml-get-attribute child 'xmlns)
+		 "urn:ietf:params:xml:ns:xmpp-stanzas")
+	(throw 'condition (jabber-xml-node-name child))))))
+
 (defvar jabber-stream-error-messages
   (list
    (cons 'bad-format "Bad XML format")
@@ -423,21 +464,66 @@ See section 9.3 of XMPP Core."
   (signal 'jabber-error
 	  (list error-type condition text app-specific)))
 
-(defun jabber-play-sound-file (soundfile)
-  (if (not *jabber-sound-playing*)
-      (progn
-	(setq *jabber-sound-playing* t)
-	(run-with-idle-timer 0.01 nil 
-			     #'(lambda (sf)
-			       (condition-case nil
-				   ;; play-sound-file might display "Could not set sample rate" in
-				   ;; echo area.  Don't let this erase the previous message.
-				   (let ((old-message (current-message)))
-				     (play-sound-file sf)
-				     (setq *jabber-sound-playing* nil)
-				     (message "%s" old-message))
-				 (error (setq *jabber-sound-playing* nil))))
-			     soundfile))))
+(defun jabber-unhex (string)
+  "Convert a hex-encoded UTF-8 string to Emacs representation.
+For example, \"ji%C5%99i@%C4%8Dechy.example/v%20Praze\" becomes
+\"jiři@čechy.example/v Praze\"."
+  (decode-coding-string (url-unhex-string string) 'utf-8))
+
+(defun jabber-handle-uri (uri &rest ignored-args)
+  "Handle XMPP links according to draft-saintandre-xmpp-iri-04.
+See Info node `(jabber)XMPP URIs'."
+  (interactive "sEnter XMPP URI: ")
+
+  (when (string-match "//" uri)
+    (error "URIs with authority part are not supported"))
+
+  ;; This regexp handles three cases:
+  ;; xmpp:romeo@montague.net
+  ;; xmpp:romeo@montague.net?roster
+  ;; xmpp:romeo@montague.net?roster;name=Romeo%20Montague;group=Lovers
+  (unless (string-match "^xmpp:\\([^?]+\\)\\(\\?\\([a-z]+\\)\\(;\\(.*\\)\\)?\\)?" uri)
+    (error "Invalid XMPP URI '%s'" uri))
+
+  ;; We start by raising the Emacs frame.
+  (raise-frame)
+
+  (let ((jid (jabber-unhex (match-string 1 uri)))
+	(method (match-string 3 uri))
+	(args (let ((text (match-string 5 uri)))
+		;; If there are arguments...
+		(when text
+		  ;; ...split the pairs by ';'...
+		  (let ((pairs (split-string text ";")))
+		    (mapcar (lambda (pair)
+			      ;; ...and split keys from values by '='.
+			      (destructuring-bind (key value) 
+				  (split-string pair "=")
+				;; Values can be hex-coded.
+				(cons key (jabber-unhex value))))
+			    pairs))))))
+    ;; The full list of methods is at
+    ;; <URL:http://www.jabber.org/registrar/querytypes.html>.
+    (cond
+     ;; Join an MUC.
+     ((string= method "join")
+      (jabber-groupchat-join jid (jabber-muc-read-my-nickname jid) t))
+     ;; Register with a service.
+     ((string= method "register")
+      (jabber-get-register jid))
+     ;; Run an ad-hoc command
+     ((string= method "command")
+      ;; XXX: does the 'action' attribute make sense?
+      (jabber-ahc-execute-command jid (cdr (assoc "node" args))))
+     ;; Everything else: open a chat buffer.
+     (t
+      (jabber-chat-with jid)))))
+
+(defun url-xmpp (url)
+  "Handle XMPP URLs from internal Emacs functions."
+  ;; XXX: This parsing roundtrip is redundant, and the parser of the
+  ;; url package might lose information.
+  (jabber-handle-uri (url-recreate-url url)))  
 
 (provide 'jabber-util)
 
